"""
Module contains settings class for luna-backport3
"""

from typing import Optional

from ..settings.classes import (
    DBSetting as _DBSetting,
    ImageStoreAddressSettings as _ImageStoreAddressSettings,
    RedisDBSetting as _RedisDBSetting,
    ServiceAddressSettings as _ServiceAddressSettings,
    ServiceTimeoutsSettings as _ServiceTimeoutsSettings,
)
from .base_python_service_settings import BasePythonServiceSettingsClass


class SettingsBackport3(BasePythonServiceSettingsClass):
    """
    Settings class for luna-backport3
    Attributes:
        apiAddress(ServiceAddressSettings): luna-api address settings
        apiTimeouts(ServiceTimeoutsSettings): luna-api timeouts settings
        portraitsStorage(ImageStoreAddressSettings): samples storage settings
        portraitsStorageTimeouts(ServiceTimeoutsSettings): samples storage timeouts settings
        db(DBSetting): backport3 database settings
        storageTime(str): storage time setting
        useSamplesAsPortraits (bool): use samples as portraits or not
        maxCandidateCount (int): max candidate count into a response on matching request
        storeEvents (bool): whether to store events
        eventsDB(RedisDBSetting): redis database settings
        enableWS (bool): Whether to enable ws events support
        enablePortraits (bool): Whether to enable portraits support
        enableHandlers (bool): Whether to enable handlers support
    """

    serviceName = "luna-backport3"

    def __init__(self):
        super().__init__()
        self.db: Optional[_DBSetting] = None
        self.apiAddress: Optional[_ServiceAddressSettings] = None
        self.apiTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.portraitsStorage: Optional[_ImageStoreAddressSettings] = None
        self.portraitsStorageTimeouts: Optional[_ServiceTimeoutsSettings] = None
        self.storageTime: Optional[str] = None
        self.useSamplesAsPortraits: Optional[bool] = None
        self.maxCandidateCount: Optional[int] = None
        self.storeEvents: Optional[bool] = None
        self.eventsDB: Optional[_RedisDBSetting] = None
        self.enableWS: Optional[bool] = None
        self.enablePortraits: Optional[bool] = None
        self.enableHandlers: Optional[bool] = None

    def applySettings(self):
        """
        Deserialize loaded settings to classes
        """
        super().applySettings()
        self.db = _DBSetting(self.serviceName, self["LUNA_BACKPORT3_DB"])
        self.apiAddress = _ServiceAddressSettings(self.serviceName, "LUNA_API_ADDRESS", self["LUNA_API_ADDRESS"])
        self.apiTimeouts = _ServiceTimeoutsSettings(self.serviceName, "LUNA_API_TIMEOUTS", self["LUNA_API_TIMEOUTS"])
        self.portraitsStorage = _ImageStoreAddressSettings(
            self.serviceName,
            "LUNA_IMAGE_STORE_PORTRAITS_ADDRESS",
            self["LUNA_IMAGE_STORE_PORTRAITS_ADDRESS"],
        )
        self.portraitsStorageTimeouts = _ServiceTimeoutsSettings(
            self.serviceName, "LUNA_IMAGE_STORE_PORTRAITS_TIMEOUTS", self["LUNA_IMAGE_STORE_PORTRAITS_TIMEOUTS"]
        )
        self.storageTime = self["STORAGE_TIME"]
        self.useSamplesAsPortraits = self["USE_SAMPLES_AS_PORTRAITS"]
        self.maxCandidateCount = self["MAX_CANDIDATE_IN_RESPONSE"]
        self.storeEvents = self["LUNA_BACKPORT3_STORE_EVENTS"]
        self.eventsDB = _RedisDBSetting(self.serviceName, self["BACKPORT3_EVENTS_DB_ADDRESS"])
        self.enableWS = self["BACKPORT3_ENABLE_WS_EVENTS"]
        self.enablePortraits = self["BACKPORT3_ENABLE_PORTRAITS"]
        self.enableHandlers = self["LUNA_BACKPORT3_USE_HANDLERS"]
