from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from types import UnionType

import ujson

projectName = Path(__file__).parent.parent.parent.stem

if projectName == "luna_tasks":
    from crutches_on_wheels.cow.enums.tasks import SubTaskStatus, TaskStatus, TaskType
    from crutches_on_wheels.cow.errors.errors import ErrorInfo
else:
    from cow.enums.tasks import SubTaskStatus, TaskStatus, TaskType
    from cow.errors.errors import ErrorInfo


@dataclass
class _BaseRedisStructure:
    """Base redis structure"""

    # luna request id
    lRId: str

    def dump(self) -> str:
        """
        Dump structure
        """
        kwargs = {}
        for key, value in self.__dict__.items():
            if isinstance(value, list) and len(value) and isinstance(value[0], _BaseRedisStructure):
                kwargs[key] = [row.dump() for row in value]
            elif isinstance(value, _BaseRedisStructure):
                kwargs[key] = value.dump()
            elif isinstance(value, Enum):
                kwargs[key] = value.value
            elif hasattr(value, "asDict"):
                kwargs[key] = value.asDict()
            else:
                kwargs[key] = value
        return ujson.dumps(kwargs)

    @classmethod
    def load(cls, data: bytes | dict | str):
        """
        Init structure from bytes/dict/str
        Args:
            data: data from redis
        """
        if isinstance(data, bytes):
            data = ujson.loads(data.decode())
        elif isinstance(data, str):
            data = ujson.loads(data)
        kwargs = {}
        for key, value in data.items():
            if key not in cls.__dataclass_fields__:
                raise RuntimeError(f"{key} not in {cls.__dataclass_fields__}")
            if cls.__dataclass_fields__[key].type.__class__ is not UnionType and issubclass(
                cls.__dataclass_fields__[key].type, Enum
            ):
                kwargs[key] = cls.__dataclass_fields__[key].type(value)
            elif hasattr(cls.__dataclass_fields__[key].type, "fromDict"):
                kwargs[key] = cls.__dataclass_fields__[key].type.fromDict(value)
            else:
                kwargs[key] = value
        return cls(**kwargs)


@dataclass
class TaskContentRequest(_BaseRedisStructure):
    """Task content request"""

    accountId: str
    redisTaskId: str
    taskType: TaskType
    content: dict


@dataclass
class SubTasksContentReply(_BaseRedisStructure):
    """Sub tasks contents reply"""

    redisTaskId: str
    taskType: TaskType
    contents: list[dict]


@dataclass
class QueueSubTask(_BaseRedisStructure):
    """Sub task to process"""

    taskId: int
    subTaskId: int
    content: dict
    accountId: str
    taskType: TaskType
    startTime: float
    resultStoragePolicy: dict


@dataclass
class QueueSubTasks(_BaseRedisStructure):
    """Sub tasks to process"""

    taskId: int
    subTaskIds: list[int]
    contents: list[dict]
    accountId: str
    taskType: TaskType
    startTime: float
    resultStoragePolicy: dict

    def dump(self) -> dict:
        return {
            ujson.dumps(
                {
                    "accountId": self.accountId,
                    "taskId": self.taskId,
                    "subTaskId": subTaskId,
                    "content": content,
                    "taskType": self.taskType.value,
                    "lRId": self.lRId,
                    "startTime": self.startTime,
                    "resultStoragePolicy": self.resultStoragePolicy,
                }
            ): self.taskId
            for subTaskId, content in zip(self.subTaskIds, self.contents)
        }


@dataclass
class SubTaskResult(_BaseRedisStructure):
    """Sub task result"""

    taskId: int
    subTaskId: int
    resultId: str
    status: SubTaskStatus
    taskType: TaskType
    accountId: str


@dataclass
class ResultAggregatorSubTask(_BaseRedisStructure):
    """Result aggregator sub task"""

    accountId: str
    taskId: int
    subTaskResults: list[SubTaskResult]
    taskType: TaskType
    taskErrors: list[dict]
    storeResults: bool
    resultStoragePolicy: dict

    @classmethod
    def load(cls, data: bytes | dict | str):
        if isinstance(data, bytes):
            data = ujson.loads(data.decode())
        elif isinstance(data, str):
            data = ujson.loads(data)
        return ResultAggregatorSubTask(
            accountId=data["accountId"],
            taskId=data["taskId"],
            subTaskResults=[SubTaskResult.load(s) for s in data["subTaskResults"]],
            taskType=TaskType(data["taskType"]),
            taskErrors=data["taskErrors"],
            storeResults=data["storeResults"],
            lRId=data["lRId"],
            resultStoragePolicy=data["resultStoragePolicy"],
        )


@dataclass
class TaskResult(_BaseRedisStructure):
    """Task result"""

    taskId: int
    status: TaskStatus
    taskType: TaskType
    resultId: str | None = None


@dataclass
class SubTaskError(_BaseRedisStructure):
    """Sub task error"""

    taskId: int
    taskType: TaskType
    error: ErrorInfo
    subTaskId: int
    additionalInfo: str | None = None


@dataclass
class SubTaskStart(_BaseRedisStructure):
    """Sub task start message"""

    taskId: int
    taskType: TaskType
    subTaskId: int


@dataclass
class SubTaskCancel(_BaseRedisStructure):
    """Sub task/collect result cancel (on shutdown) message"""

    taskType: TaskType
    subTaskId: int
    taskId: int
